from signal import SIGTERM, signal
from string import ascii_letters, digits
from time import time

from boba_fetch.rpc import _random_string, UnixSocketRPCServer


def test_random_string() -> None:
    for _ in range(1000):
        assert len(_random_string(32)) == 32
        assert all(c in (ascii_letters + digits) for c in _random_string(32))


def test_rpc_server() -> None:
    class TestServer(UnixSocketRPCServer):
        def __init__(self, msg: str) -> None:
            self.msg = msg
            UnixSocketRPCServer.__init__(self)

        @UnixSocketRPCServer.serve
        def get_msg(self) -> str:
            return self.msg

    with TestServer('stuff').proxy() as proxy:
        assert proxy.get_msg() == 'stuff'


def test_uncooperative_shutdown() -> None:
    class UncooperativeTestServer(UnixSocketRPCServer):
        def __init__(self, msg: str) -> None:
            self.msg = msg
            UnixSocketRPCServer.__init__(self)

        @UnixSocketRPCServer.serve
        def setup(self) -> None:
            def sigterm_ignore(_, __):
                pass
            signal(SIGTERM, sigterm_ignore)

        @UnixSocketRPCServer.serve
        def stuff(self) -> str:
            return 'stuff'

    with UncooperativeTestServer('stuff').proxy() as proxy:
        proxy.setup()
        shutdown_start = time()

    assert time() - shutdown_start <= 2.0



from collections import defaultdict
from contextlib import contextmanager
from logging import DEBUG, INFO  # noqa: F401 # pylint: disable=W0611
from threading import current_thread


from logzero import loglevel as set_loglevel  # noqa: F401 # pylint: disable=W0611
from logzero import logger as log  # noqa: F401 # pylint: disable=W0611
from logzero import LogFormatter, formatter as set_formatter

DEFAULT_FMT = '%(color)s[%(asctime)s] [%(process)d] [%(levelname)s] [%(threadName)s]%(end_color)s %(message)s'
DEBUG_FMT = ('%(color)s[%(asctime)s] [%(process)d] [%(levelname)s] [%(threadName)s] [%(processName)s:%(module)s:'
             '%(funcName)s:%(lineno)d]%(end_color)s %(message)s')


class DefaultLogFormatter(LogFormatter):
    def __init__(self, formatters,
                 default_format=DEFAULT_FMT,
                 date_format='%Y-%m-%d %H:%M:%S %z'):
        self.formats = defaultdict(lambda: default_format)
        for key, value in formatters.items():
            self.formats[key] = value
        LogFormatter.__init__(self, fmt=default_format, datefmt=date_format)

    def format(self, record):
        self._fmt = self.formats[record.levelno]
        return LogFormatter.format(self, record)


@contextmanager
def thread_name(new_name: str):
    thread = current_thread()
    original_name = thread.name
    thread.name = new_name
    try:
        yield
    finally:
        thread.name = original_name


set_formatter(DefaultLogFormatter({DEBUG: DEBUG_FMT},
                                  default_format=DEFAULT_FMT))



from dataclasses import dataclass
from multiprocessing import Process, Event as event, Queue
from multiprocessing.synchronize import Event
from pathlib import Path
from pkg_resources import get_distribution
from signal import SIGTERM, SIGINT, signal
from tempfile import mkdtemp
from typing import Optional, List
from warnings import simplefilter

import click
from click_path import PathlibPath
import cryptography

from boba_fetch.log import log, thread_name, set_loglevel, DEBUG, INFO
from boba_fetch.poller import RemotePoller
from boba_fetch.rpc import UnixSocketRPCClient
from boba_fetch.syncer import Syncer, WatchParams
from boba_fetch.tracker import Tracker


simplefilter('ignore', cryptography.utils.CryptographyDeprecationWarning)


@dataclass
class BobaConfig:
    remote_host: str
    remote_dir: Path
    local_dir: Path
    staging_dir: Path
    tracking_dir: Path
    num_sync_workers: int
    delete_src: bool
    execute_on_complete: Optional[str]
    polling_delay: float

    def to_watch_params(self, queue: Queue) -> WatchParams:
        return WatchParams(self.remote_host, self.local_dir, queue, self.delete_src,
                           self.execute_on_complete)


def run_poller(config: BobaConfig,
               tracker: UnixSocketRPCClient,
               queue: Queue,
               halt_flag: Event):
    with thread_name("Poller"):
        poller = RemotePoller(config.remote_host, tracker)
        poller.poll(config.polling_delay, config.remote_dir, queue, halt_flag)


def start_poller(config: BobaConfig,
                 tracker: UnixSocketRPCClient,
                 queue: Queue,
                 halt_flag: Event) -> Process:
    poller_proc = Process(target=run_poller, args=(config, tracker, queue, halt_flag))
    poller_proc.start()
    return poller_proc


def run_syncer(config: BobaConfig,
               tracker: UnixSocketRPCClient,
               queue: Queue,
               halt_flag: Event,
               index: int):
    with thread_name(f'Syncer {index}'):
        syncer = Syncer(tracker, halt_flag, config.staging_dir)
        syncer.watch_queue(config.to_watch_params(queue))


def start_syncers(config: BobaConfig,
                  tracker: UnixSocketRPCClient,
                  queue: Queue,
                  halt_flag: Event) -> List[Process]:
    syncer_procs = [Process(target=run_syncer,
                            args=(config, tracker,
                                  queue, halt_flag, i)) for i in range(config.num_sync_workers)]
    for proc in syncer_procs:
        proc.start()
    return syncer_procs


@click.command()
@click.version_option(get_distribution('boba-fetch').version)
@click.option('-r', '--remote', required=True,
              help='The remote host to poll for file updates. Must be SSH accessible via public key '
                   'and configured in your ~/.ssh/config')
@click.option('-s', '--src-dir', required=True, type=PathlibPath(exists=False),
              help='The directory on the remote host to check for files to download')
@click.option('-d', '--dst-dir', required=True, type=PathlibPath(exists=True, file_okay=False,
                                                                 dir_okay=True, writable=True,
                                                                 resolve_path=True),
              help='Local directory to sync the files to from the remote host')
@click.option('-g', '--staging-dir', default=None, type=PathlibPath(exists=True, file_okay=False,
                                                                    dir_okay=True, writable=True,
                                                                    resolve_path=True),
              help='Staging directory where files are saved, defaults to randomized temporary directory')
@click.option('-t', '--tracking-dir', required=True, type=PathlibPath(exists=True, file_okay=False,
                                                                      dir_okay=True, writable=True,
                                                                      resolve_path=True),
              help='Tracking directory where we keep track of files that have been synced')
@click.option('-n', '--num-sync-workers', default=3, type=int,
              help='Number of sync workers to use for downloads, defaults to 3')
@click.option('--delete-src', is_flag=True, default=False, help='Indicates to whether to delete the source files on '
                                                                'the remote system')
@click.option('-e', '--execute-on-complete', default=None, help='A command to execute every time a sync is complete')
@click.option('-p', '--polling-delay', default=30.0, type=float,
              help='Number of seconds to wait between polls of the remote directory, defaults to 30')
@click.option('--debug', is_flag=True, default=False, help='Turn on debug logging')
def main(remote: str,
         src_dir: Path,
         dst_dir: Path,
         staging_dir: Optional[Path],
         tracking_dir: Path,
         num_sync_workers: int,
         delete_src: bool,
         execute_on_complete: Optional[str],
         polling_delay: float,
         debug: bool):
    set_loglevel(DEBUG if debug else INFO)
    queue: Queue = Queue()
    halt_flag = event()
    config = BobaConfig(remote, src_dir, dst_dir, staging_dir or Path(mkdtemp()),
                        tracking_dir, num_sync_workers, delete_src, execute_on_complete,
                        polling_delay)
    with Tracker(config.tracking_dir).proxy() as tracker:
        syncer_procs = start_syncers(config, tracker, queue, halt_flag)
        poller_proc = start_poller(config, tracker, queue, halt_flag)

        def handle_signal(_, __):
            halt_flag.set()
        signal(SIGTERM, handle_signal)
        signal(SIGINT, handle_signal)
        poller_proc.join()
        log.info("Poller shutdown complete.")
        for sync_proc in syncer_procs:
            sync_proc.join()
        log.info("Syncer shutdown complete.")
        log.info("Done.")



from multiprocessing import Queue
from pathlib import Path
from subprocess import SubprocessError
from threading import Thread, Event
from time import sleep
from typing import List

from pytest import raises, mark
from pytest_mock import MockerFixture

from boba_fetch.syncer import SyncParams, Syncer, WatchParams
from boba_fetch.tracker import Tracker

from tests.unit.utils import tempdir, generate_big_sparse_file


def test_sync_params() -> None:
    params = SyncParams('test-host',
                        Path.cwd(),
                        Path('/remote/src'),
                        delete=False,
                        execute_on_complete='echo stuff')
    assert params.on_complete == ['echo', 'stuff']
    assert params.name == 'src'


def test_sync_params_from_watch() -> None:
    params = WatchParams('test-host',
                         Path.cwd(),
                         Queue(),
                         delete=False,
                         execute_on_complete='echo stuff')
    sync_params = SyncParams.from_watch(params, Path('/remote/src'))
    assert sync_params.on_complete == ['echo', 'stuff']
    assert sync_params.remote_path == Path('/remote/src')


def test_syncer_init() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker)
            assert syncer.tracker is tracker


def test_syncer_non_existent_temp_dir() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            with raises(ValueError, match=r'Temporary directory .*/\.temp does not exist'):
                Syncer(tracker, temp_dir=(tmp_dir / '.temp'))


def test_syncer_no_read_temp_dir() -> None:
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            temp_dir = (tmp_dir / '.temp')
            temp_dir.mkdir()
            temp_dir.chmod(0o200)
            with raises(ValueError, match=r'No read-write access to temporary directory: .*/\.temp'):
                Syncer(tracker, temp_dir=temp_dir)


@mark.parametrize('delete,temp_dir,expected', [
    (False, None, ['rsync', '-Pparvzy', '--chmod=u=rw,g=r,o=r', 'remote:"/src"', '/dst']),
    (True, None, ['rsync', '-Pparvzy', '--chmod=u=rw,g=r,o=r', '--remove-source-files', 'remote:"/src"', '/dst']),
    (True, Path('/tmp'), ['rsync', '-Pparvzy', '--chmod=u=rw,g=r,o=r', '-T', '/tmp', '--remove-source-files',
     'remote:"/src"', '/dst']),
])
def test_cmd_generation(delete, temp_dir, expected):
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker, temp_dir=temp_dir)
            params = SyncParams('remote', '/dst', '/src', delete)
            assert syncer.cmd(params) == expected


def local_mock_cmd(self, params: SyncParams) -> List[str]:
    result = ['rsync', '-Pparvzy', '--chmod=u=rwX,g=rX,o=rX']
    if self.temp_dir:
        result += ['-T', str(self.temp_dir)]
    if params.delete:
        result += ['--remove-source-files']
    return result + [f'{str(params.remote_path)}', str(params.local_dir)]


_1GiB = 1024 * 1024 * 1024
_200MiB = 1024 * 1024 * 200


def test_sync_down_basic(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file = generate_big_sparse_file(src_dir / 'src.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        dst_file = dst_dir / 'src.file'
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker, temp_dir=stg_dir)
            params = SyncParams('remote', dst_dir, src_file, delete=False)
            assert syncer.sync_down(params)
            assert dst_file.is_file()
            assert dst_file.stat().st_size == _200MiB


def test_sync_down_terminate(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file = generate_big_sparse_file(src_dir / 'src.file', _1GiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        dst_file = dst_dir / 'src.file'
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            halt_flag = Event()
            syncer = Syncer(tracker, temp_dir=stg_dir, halt_flag=halt_flag)
            params = SyncParams('remote', dst_dir, src_file, delete=False)

            def run():
                assert not syncer.sync_down(params)
            thread = Thread(target=run)
            thread.start()
            sleep(1)
            halt_flag.set()
            thread.join(1)
            assert not thread.is_alive()
            assert dst_file.is_file()
            assert dst_file.stat().st_size < _1GiB


def test_sync_down_no_start(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file = generate_big_sparse_file(src_dir / 'src.file', _1GiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        dst_file = dst_dir / 'src.file'
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            halt_flag = Event()
            halt_flag.set()
            syncer = Syncer(tracker, temp_dir=stg_dir, halt_flag=halt_flag)
            params = SyncParams('remote', dst_dir, src_file, delete=False)

        def run():
            assert not syncer.sync_down(params)
        thread = Thread(target=run)
        thread.start()
        thread.join(1)
        assert not thread.is_alive()
        assert not dst_file.is_file()


def test_sync_down_directory(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_subdir = src_dir / 'stuff'
        src_subdir.mkdir()
        generate_big_sparse_file(src_subdir / 'A.file', _200MiB)
        generate_big_sparse_file(src_subdir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        dst_file1 = dst_dir / 'stuff' / 'A.file'
        dst_file2 = dst_dir / 'stuff' / 'B.file'
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker, temp_dir=stg_dir)
            params = SyncParams('remote', dst_dir, src_subdir, delete=False)
            assert syncer.sync_down(params)
            assert dst_file1.is_file()
            assert dst_file1.stat().st_size == _200MiB
            assert dst_file2.is_file()
            assert dst_file2.stat().st_size == _200MiB
            assert dst_dir.stat().st_mode == 0o40755
            assert dst_file1.stat().st_mode == 0o100644
            assert dst_file2.stat().st_mode == 0o100644


def test_sync_down_directory_and_delete(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_subdir = src_dir / 'stuff'
        src_subdir.mkdir()
        src_file_1 = generate_big_sparse_file(src_subdir / 'A.file', _200MiB)
        src_file_2 = generate_big_sparse_file(src_subdir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        dst_file1 = dst_dir / 'stuff' / 'A.file'
        dst_file2 = dst_dir / 'stuff' / 'B.file'
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker, temp_dir=stg_dir)
            params = SyncParams('remote', dst_dir, src_subdir, delete=True)
            assert syncer.sync_down(params)
            assert dst_file1.is_file()
            assert dst_file1.stat().st_size == _200MiB
            assert dst_file2.is_file()
            assert dst_file2.stat().st_size == _200MiB
            assert dst_dir.stat().st_mode == 0o40755
            assert dst_file1.stat().st_mode == 0o100644
            assert dst_file2.stat().st_mode == 0o100644
            assert not src_file_1.is_file()
            assert not src_file_2.is_file()


def test_on_complete():
    params = SyncParams('test-host',
                        Path.cwd(),
                        Path('/remote/src'),
                        delete=False,
                        execute_on_complete='echo stuff')
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker)
            syncer.execute_on_complete(params)


def test_on_complete_exception(mocker: MockerFixture):
    def error_mock_split(cmd) -> None:
        print(cmd)
        raise SubprocessError('shit happens')
    mocker.patch('shlex.split', error_mock_split)
    params = SyncParams('test-host',
                        Path.cwd(),
                        Path('/remote/src'),
                        delete=False,
                        execute_on_complete='echo stuff')
    with tempdir() as tmp_dir:
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            syncer = Syncer(tracker)
            syncer.execute_on_complete(params)


def test_watch_queue_basic(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file_1 = generate_big_sparse_file(src_dir / 'A.file', _200MiB)
        src_file_2 = generate_big_sparse_file(src_dir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        queue: Queue = Queue()
        queue.put(src_file_1)
        queue.put(src_file_2)
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            params = WatchParams('test-host',
                                 dst_dir,
                                 queue,
                                 delete=False)
            syncer = Syncer(tracker, temp_dir=stg_dir)

            def wait_then_halt():
                sleep(4)
                syncer.halt()
            thread = Thread(target=wait_then_halt)
            thread.start()
            syncer.watch_queue(params)
            thread.join()
            assert (dst_dir / 'A.file').is_file()
            assert (dst_dir / 'A.file').stat().st_size == _200MiB
            assert (dst_dir / 'B.file').is_file()
            assert (dst_dir / 'B.file').stat().st_size == _200MiB
            assert tracker.completed_at('A.file') != 0.0
            assert tracker.completed_at('B.file') != 0.0


def test_watch_queue_delete(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file_1 = generate_big_sparse_file(src_dir / 'A.file', _200MiB)
        src_file_2 = generate_big_sparse_file(src_dir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        queue: Queue = Queue()
        queue.put(src_file_1)
        queue.put(src_file_2)
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            params = WatchParams('test-host',
                                 dst_dir,
                                 queue,
                                 delete=True)
            syncer = Syncer(tracker, temp_dir=stg_dir)

            def wait_then_halt():
                sleep(4)
                syncer.halt()
            thread = Thread(target=wait_then_halt)
            thread.start()
            syncer.watch_queue(params)
            thread.join()
            assert not (src_dir / 'A.file').is_file()
            assert not (src_dir / 'B.file').is_file()
            assert (dst_dir / 'A.file').is_file()
            assert (dst_dir / 'A.file').stat().st_size == _200MiB
            assert (dst_dir / 'B.file').is_file()
            assert (dst_dir / 'B.file').stat().st_size == _200MiB
            assert tracker.completed_at('A.file') != 0.0
            assert tracker.completed_at('B.file') != 0.0


def test_watch_queue_multiple_syncers(mocker: MockerFixture) -> None:
    mocker.patch('boba_fetch.syncer.Syncer.cmd', local_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file_1 = generate_big_sparse_file(src_dir / 'A.file', _200MiB)
        src_file_2 = generate_big_sparse_file(src_dir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        queue: Queue = Queue()
        queue.put(src_file_1)
        queue.put(src_file_1)
        queue.put(src_file_2)
        queue.put(src_file_2)
        halt = Event()
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            params = WatchParams('test-host',
                                 dst_dir,
                                 queue,
                                 delete=False)

            def run_syncer():
                syncer = Syncer(tracker, temp_dir=stg_dir, halt_flag=halt)
                syncer.watch_queue(params)
            thread1 = Thread(target=run_syncer)
            thread1.start()
            thread2 = Thread(target=run_syncer)
            thread2.start()
            sleep(5)
            halt.set()
            thread1.join()
            thread2.join()
            assert (dst_dir / 'A.file').is_file()
            assert (dst_dir / 'A.file').stat().st_size == _200MiB
            assert (dst_dir / 'B.file').is_file()
            assert (dst_dir / 'B.file').stat().st_size == _200MiB
            assert tracker.completed_at('A.file') != 0.0
            assert tracker.completed_at('B.file') != 0.0


def test_watch_queue_error(mocker: MockerFixture) -> None:
    def error_mock_cmd(self, params: SyncParams) -> List[str]:
        raise RuntimeError('shit happens')
    mocker.patch('boba_fetch.syncer.Syncer.cmd', error_mock_cmd)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file_1 = generate_big_sparse_file(src_dir / 'A.file', _200MiB)
        src_file_2 = generate_big_sparse_file(src_dir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        queue: Queue = Queue()
        queue.put(src_file_1)
        queue.put(src_file_2)
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            params = WatchParams('test-host',
                                 dst_dir,
                                 queue,
                                 delete=False)
            syncer = Syncer(tracker, temp_dir=stg_dir)

            def wait_then_halt():
                sleep(1)
                syncer.halt()
            thread = Thread(target=wait_then_halt)
            thread.start()
            syncer.watch_queue(params)
            thread.join()
            assert tracker.completed_at('A.file') == 0.0
            assert tracker.completed_at('B.file') == 0.0


def test_watch_queue_sync_fail(mocker: MockerFixture) -> None:
    def mock_sync_down_false(_, __) -> bool:
        return False
    mocker.patch('boba_fetch.syncer.Syncer.sync_down', mock_sync_down_false)
    with tempdir() as tmp_dir:
        src_dir = tmp_dir / 'src'
        src_dir.mkdir()
        src_file_1 = generate_big_sparse_file(src_dir / 'A.file', _200MiB)
        src_file_2 = generate_big_sparse_file(src_dir / 'B.file', _200MiB)
        dst_dir = tmp_dir / 'dst'
        dst_dir.mkdir()
        stg_dir = tmp_dir / 'stg'
        stg_dir.mkdir()
        queue: Queue = Queue()
        queue.put(src_file_1)
        queue.put(src_file_2)
        with Tracker(tmp_dir / '.track').proxy() as tracker:
            params = WatchParams('test-host',
                                 dst_dir,
                                 queue,
                                 delete=False)
            syncer = Syncer(tracker, temp_dir=stg_dir)

            def wait_then_halt():
                sleep(1)
                syncer.halt()
            thread = Thread(target=wait_then_halt)
            thread.start()
            syncer.watch_queue(params)
            thread.join()
            assert tracker.completed_at('A.file') == 0.0
            assert tracker.completed_at('B.file') == 0.0
